import planktofeatures.extractors
import os
import libifcb
import time
import csv

class MainJob:
    def __init__(self, options, progress_reporting_function = lambda prop, etr : print(str(int(prop*10000)/100) + "% done - ETR " + str(int(etr)) + "s"), log_function = lambda txt : print("[LOG] " + txt), error_function = lambda txt : print("[ERR] " + txt)):
        self.options = options
        self.report_progress = progress_reporting_function
        self.log_function = log_function
        self.error_function = error_function

        input_files_list = []
        for input_file_path in options["input_files"]:
            input_files_list.append(os.path.abspath(input_file_path))

        ifcb_bins = []
        ifcb_files = []
        roi_readers = []

        intermediate_files_list = set()
        for file_name in input_files_list:
            intermediate_files_list.add(os.path.splitext(file_name)[0])

        for file_name in intermediate_files_list:
            ifcb_bins.append(os.path.basename(file_name))
            ifcb_files.append(file_name)


        self.log_function("Creating ROI readers")

        for i in range(len(ifcb_files)):
            self.log_function("Loading \"" + ifcb_files[i] + ".hdr\"")
            roi_readers.append(libifcb.ROIReader(ifcb_files[i] + ".hdr", ifcb_files[i] + ".adc", ifcb_files[i] + ".roi"))

        self.roi_readers = roi_readers
        self.ifcb_files = ifcb_files
        self.ifcb_bins = ifcb_bins
        self.total_rois = 0
        for roi_reader in roi_readers:
            self.total_rois += len(roi_reader.rows)

    def generate_features_one_file(self, sample, source_file, bin_id, csv_file):
        self.log_function("Processing \"" + source_file + ".roi\"")
        first = True
        feature_extractor = planktofeatures.extractors.WHOIVersion4()
        force_overwrite = False
        ignore_existing = False
        if "force_overwrite" in self.options:
            force_overwrite = self.options["force_overwrite"]

        if "ignore_existing" in self.options:
            ignore_existing = self.options["ignore_existing"]
        if os.path.isfile(csv_file):
            if force_overwrite:
                self.log_function("File \"" + csv_file + "\" already exists, overwriting anyway as requested")
            else:
                if ignore_existing:
                    self.log_function("File \"" + csv_file + "\" already exists, ignoring as requested")
                    return None
                else:
                    raise RuntimeError("File \"" + csv_file + "\" already exists, use --ignore to surpress this error, and --force to overwrite files")
        with open(csv_file, "w", newline="") as csvfile:
            csv_writer = csv.writer(csvfile, delimiter=",", quotechar="\"", quoting=csv.QUOTE_MINIMAL)
            adc_row_idx = 0
            total_rows = len(sample.rows)
            for adc_row_obj in sample.rows:
                adc_row_idx += 1
                self.currently_processing += 1
                img = adc_row_obj.image
                if img is not None:
                    feature_object = feature_extractor.process(img)
                    row_features = feature_object.values
                    observation_id = bin_id + "_" + str(adc_row_obj.index).zfill(5)
                    if first:
                        first = False
                        akeys = list(map(str.lower,row_features.keys()))
                        csv_writer.writerow(["roi_number", "roi_id", "feature_extractor", *akeys])
                    csv_writer.writerow([adc_row_idx, observation_id,"whoi_v4", *row_features.values()])

                if adc_row_idx % 16 == 0:
                    self.last_time = time.time()
                    proportion = self.currently_processing/self.total_rois
                    elapsed_time = self.last_time - self.first_time
                    time_per_roi = elapsed_time / self.currently_processing
                    remaining_rois = self.total_rois - self.currently_processing
                    remaining_time = time_per_roi * remaining_rois
                    self.report_progress(proportion, remaining_time) # Placeholdr

    def execute(self):
        self.first_time = time.time()
        self.currently_processing = 0
        self.last_time = time.time()

        os.makedirs(self.options["output_folder"], exist_ok=True)
        for roi_reader_idx in range(len(self.roi_readers)):
            csv_file = os.path.join(self.options["output_folder"], self.ifcb_bins[roi_reader_idx] + "_features_v4.csv")
            self.generate_features_one_file(self.roi_readers[roi_reader_idx], self.ifcb_files[roi_reader_idx], self.ifcb_bins[roi_reader_idx], csv_file)

        if len(self.roi_readers) == 0:
            self.log_function("No ROI files loaded! - Are you sure you specified the right input files?")
        else:
            self.log_function("Processed " + str(len(self.roi_readers)) + " IFCB bins")



import os
import libifcb
from datetime import datetime, timezone
import csv
import json
import re
import zipfile
import io
import time

csv.field_size_limit(1024 * 1024 * 1024)

class EcoTaxaTSVReader:
    def __init__(self, tsv_fh):
        self.tsv_fh = tsv_fh
        self.tsv_fh.seek(0, os.SEEK_END)
        self.tsv_total_size = self.tsv_fh.tell()
        self.tsv_fh.seek(0)
        self.tsv_reader = csv.reader(self.tsv_fh, delimiter='\t', quotechar='"')
        self.column_names = next(self.tsv_reader)
        self.column_types = next(self.tsv_reader)
        self.last_row = next(self.tsv_reader)
        self.last_row_index = 0
        self.last_row_offset = self.tsv_fh.tell()

    def __iter__(self):
        return self

    def __next__(self):
        if self.last_row is None:
            raise StopIteration()

        parsed_row = self.last_row

        try:
            self.last_row_index += 1
            self.last_row = next(self.tsv_reader)
            self.last_row_offset = self.tsv_fh.tell()
        except StopIteration:
            self.last_row = None

        return parsed_row


class EcoTaxaDump:
    def __init__(self, tsv_list, other_files_list):
        self.tsvs = []
        self.other_files = {}
        for tsv_path in tsv_list:
            self.tsvs.append(EcoTaxaTSVReader(open(tsv_path, "r")))

        print(next(self.tsvs[0]))

class MainJob:
    def calc_progress_report(self):
        self.last_time = time.time()
        proportion = self.currently_processing/self.total_rois
        elapsed_time = self.last_time - self.first_time
        time_per_roi = elapsed_time / self.currently_processing
        remaining_rois = self.total_rois - self.currently_processing
        remaining_time = time_per_roi * remaining_rois
        self.report_progress(proportion, remaining_time)

    def __init__(self, options, progress_reporting_function = lambda prop, etr : print(str(int(prop*10000)/100) + "% done - ETR " + str(int(etr)) + "s"), log_function = lambda txt : print("[LOG] " + txt), error_function = lambda txt : print("[ERR] " + txt)):

        self.options = options
        self.report_progress = progress_reporting_function
        self.log_function = log_function
        self.error_function = error_function

        #print(options)

        if "data_file" in options.keys():
            self.with_images = True
        else:
            self.with_images = False

        input_files_list = []
        for input_file_path in options["input_files"]:
            input_files_list.append(os.path.abspath(input_file_path))



        #intermediate_files_list = set()
        tsv_files_list = set()
        other_files_list = set()
        for file_name in input_files_list:
            splitext = os.path.splitext(file_name)
            if (splitext[1] == ".tsv"):
                tsv_files_list.add(file_name)
            else:
                other_files_list.add(file_name)

        tsv_files_list = list(tsv_files_list)
        other_files_list = list(other_files_list)
        if len(tsv_files_list) == 0:
            raise RuntimeException("No TSV files supplied!")

        self.eco_taxa_dump_obj = EcoTaxaDump(tsv_files_list, other_files_list)

    def execute(self):
        self.first_time = time.time()
        self.currently_processing = 0
        self.last_time = time.time()







#!/bin/python3

# Copyright 2025, A Baldwin <alewin@noc.ac.uk>, National Oceanography Centre.
#
# This file is part of SeaSTAR, a tool for processing IFCB data.
#
# SeaSTAR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License (version 3 only)
# as published by the Free Software Foundation.
#
# SeaSTAR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v3
# along with SeaSTAR.  If not, see <http://www.gnu.org/licenses/>.

import tkinter
import tkinter.filedialog
import tkinter.messagebox
import tkinter.scrolledtext
import tkinter.font
import os

class ScrollableFrame(tkinter.Frame):
    def _on_mousewheel(self, event):
        # Linux uses Button-4 and Button-5, Windows/Mac use MouseWheel
        if event.num == 4: # Linux scroll up
            self.canvas.yview_scroll(-1, "units")
        elif event.num == 5: # Linux scroll down
            self.canvas.yview_scroll(1, "units")
        else: # Windows and MacOS
            self.canvas.yview_scroll(int(-1*(event.delta/120)), "units")

    def bind_mouse_wheel(self, widget):
        """Recursively bind mouse wheel to a widget and all its children."""
        # Support for Windows/MacOS
        widget.bind("<MouseWheel>", self._on_mousewheel)
        # Support for Linux
        widget.bind("<Button-4>", self._on_mousewheel)
        widget.bind("<Button-5>", self._on_mousewheel)

        for child in widget.winfo_children():
            self.bind_mouse_wheel(child)

    def __init__(self, container, *args, **kwargs):
        super().__init__(container, *args, **kwargs)

        # 1. Create a canvas object and a vertical scrollbar
        self.canvas = tkinter.Canvas(self)
        self.scrollbar = tkinter.Scrollbar(self, orient="vertical", command=self.canvas.yview)

        # 2. Create the frame that will hold the actual content
        self.scrollable_content = tkinter.Frame(self.canvas)

        # 3. Bind the configuration of the internal frame to the canvas scroll region
        self.scrollable_content.bind(
            "<Configure>",
            lambda e: self.canvas.configure(scrollregion=self.canvas.bbox("all"))
        )

        # 4. Create a window inside the canvas to house the internal frame
        self.canvas.create_window((0, 0), window=self.scrollable_content, anchor="nw")

        # 5. Link the canvas to the scrollbar
        self.canvas.configure(yscrollcommand=self.scrollbar.set)

        # 6. Pack everything
        self.canvas.pack(side="left", fill="both", expand=True)
        self.scrollbar.pack(side="right", fill="y")

        #self.canvas.bind_all("<MouseWheel>", lambda e: self.canvas.yview_scroll(int(-1*(e.delta/120)), "units"))

class SeaSTARGUI():
    def __set_dynamic_reflow(self, widget):
        widget.master.bind("<Configure>", self.__get_reflow_closure(widget))

    def __get_reflow_closure(self, widget):
        def closure(event):
            widget.config(wraplength=widget.master.winfo_width())
        return closure

    def __create_scrollable_standard_label(self, parent, text):
        container = tkinter.Frame(parent)

        label = tkinter.Label(container, text=text, justify="left")
        label.pack(anchor="w")
        self.__set_dynamic_reflow(label)

        return (container, label)

    def render_job_form(self, job_name):
        for widget in self.root.winfo_children():
            widget.destroy()

        self.root.columnconfigure(index=0, weight=1)
        self.root.rowconfigure(index=1, weight=1)
        self.head_text = self.__create_scrollable_standard_label(self.root, job_name)
        self.head_text[0].grid(column=0, row=0, sticky="ew", padx=self.standard_xpad * 2, pady=self.standard_ypad)

        self.main_scrollable_frame = ScrollableFrame(self.root)
        self.main_scrollable_frame.grid(column=0, row=1, sticky="nsew", padx=0, pady=0)

        for i in range(50):
            tkinter.Label(self.main_scrollable_frame.scrollable_content, text=f"This is item number {i}").pack(pady=5)

        self.main_scrollable_frame.bind_mouse_wheel(self.main_scrollable_frame.scrollable_content)

        #testtext = self.__create_scrollable_standard_label(self.main_scrollable_frame.scrollable_content, "Lorem ipsum dolor sit amet consectetur adipiscing elit. Quisque faucibus ex sapien vitae pellentesque sem placerat. In id cursus mi pretium tellus duis convallis. Tempus leo eu aenean sed diam urna tempor. Pulvinar vivamus fringilla lacus nec metus bibendum egestas. Iaculis massa nisl malesuada lacinia integer nunc posuere. Ut hendrerit semper vel class aptent taciti sociosqu. Ad litora torquent per conubia nostra inceptos himenaeos.")



    def render_start_page(self):
        self.root.columnconfigure(index=0, weight=1)
        self.head_text = self.__create_scrollable_standard_label(self.root, "Welcome to SeaSTAR! Start by selecting a job.")
        self.head_text[0].grid(column=0, row=0, sticky="ew", padx=self.standard_xpad * 2, pady=self.standard_ypad)


        def create_job_onclick_function(hlframe,io_def):
            def closure(event):
                print("Clicked! " + io_def["name"])
                self.render_job_form(io_def_key)
                #hlframe.config(relief=tkinter.SUNKEN)
            return closure

        def create_job_onenter_function(hlframe):
            def closure(event):
                hlframe.config(relief=tkinter.RAISED)
            return closure

        def create_job_onleave_function(hlframe):
            def closure(event):
                hlframe.config(relief=tkinter.GROOVE)
            return closure

        row_index = 0
        for io_def_key in self.module_io_defs.keys():
            print(io_def_key)
            io_def = self.module_io_defs[io_def_key]

            row_index += 1
            job_frame = tkinter.Frame(self.root, borderwidth=2, relief=tkinter.GROOVE)
            job_frame.grid(column=0, row=row_index, sticky="ew", padx=self.standard_xpad * 2, pady=self.standard_ypad)
            job_frame.columnconfigure(index=0, weight=1)
            job_frame.columnconfigure(index=1, weight=0)

            onclick_function = create_job_onclick_function(job_frame,io_def)
            onenter_function = create_job_onenter_function(job_frame)
            onleave_function = create_job_onleave_function(job_frame)
            job_frame.bind("<Button-1>", onclick_function)
            job_frame.bind("<Enter>", onenter_function)
            job_frame.bind("<Leave>", onleave_function)

            if "name" not in io_def.keys():
                io_def["name"] = io_def_key

            name_label = self.__create_scrollable_standard_label(job_frame, io_def["name"])
            name_label[0].grid(column=0, row=0, sticky="ew", padx=self.standard_xpad, pady=self.standard_ypad)
            name_label[0].bind("<Button-1>", onclick_function)
            name_label[0].bind("<Enter>", onenter_function)
            name_label[0].bind("<Leave>", onleave_function)
            name_label[1].bind("<Button-1>", onclick_function)
            name_label[1].bind("<Enter>", onenter_function)
            name_label[1].bind("<Leave>", onleave_function)
            name_label[1].config(font=(self.default_font["family"], self.default_font["size"], "bold"))

            if "description" not in io_def.keys():
                io_def["description"] = "This job is missing a description!"

            description_label = self.__create_scrollable_standard_label(job_frame, io_def["description"])
            description_label[0].grid(column=0, row=1, sticky="ew", padx=self.standard_xpad, pady=self.standard_ypad)
            description_label[0].bind("<Button-1>", onclick_function)
            description_label[0].bind("<Enter>", onenter_function)
            description_label[0].bind("<Leave>", onleave_function)
            description_label[1].bind("<Button-1>", onclick_function)
            description_label[1].bind("<Enter>", onenter_function)
            description_label[1].bind("<Leave>", onleave_function)

    def __init__(self, python_file_loc="", module_io_defs={}):
        self.root = tkinter.Tk()

        self.root.title("SeaSTAR")
        self.root.geometry("600x400")
        self.root.iconphoto(False, tkinter.PhotoImage(file=os.path.join(python_file_loc, "icon.png")))
        self.default_font = tkinter.font.nametofont("TkDefaultFont").actual()
        self.standard_xpad = 5
        self.standard_ypad = 5
        self.module_io_defs = module_io_defs
        self.python_file_loc = python_file_loc



        #button = tkinter.Button(self.root, text="Select files", command=button_execute_onclick)
        #button.grid(column=1, row=1, sticky="SEW", padx=standard_xpad, pady=standard_ypad)

    def enter_mainloop(self):
        self.root.mainloop()



import os
import libifcb
from datetime import datetime, timezone
import csv
import json
import re
import zipfile
import io
import time

csv.field_size_limit(1024 * 1024 * 1024)

class IFCBEntryProvider:
    def __init__(self, roi_readers, ifcb_ids, job_object, with_images = True, options = {}):
        self.roi_readers = roi_readers
        self.ifcb_ids = ifcb_ids
        self.with_images = with_images
        self.job_object = job_object
        self.reader_index = 0
        self.index = 0
        self.options = options
        self.column_sources = {}
        self.match_data_keys = {}
        self.process_time = datetime.now().replace(tzinfo=timezone.utc)

        self.mangle_nulls = False
        self.allow_short_names = False

        if "mangle_nulls" in self.options.keys():
            self.mangle_nulls = self.options["mangle_nulls"]

        if "allow_short_names" in self.options.keys():
            self.allow_short_names = self.options["allow_short_names"]

        self.ecotaxa_table_header = {
                "object_id": "[t]",
                "object_date": "[t]",
                "object_time": "[t]",
                "object_roi_width": "[f]",
                "object_roi_height": "[f]",
                "process_id": "[t]",
                "process_date": "[t]",
                "process_time": "[t]",
                "process_pixel_um": "[f]",
                "process_feature_extractor": "[t]",
                "acq_id": "[t]",
                "acq_operator": "[t]",
                "acq_instrument": "[t]",
                "acq_sn": "[t]",
                "acq_run_time": "[f]",
                "acq_grab_time_start": "[f]",
                "acq_trigger_number": "[f]",
                "acq_peak_a": "[f]",
                "acq_pmt_a": "[f]",
                "acq_peak_b": "[f]",
                "acq_pmt_b": "[f]",
                "acq_grab_time_end": "[f]",
                "acq_adc_time": "[f]",
                "acq_signal_length": "[f]",
                "acq_inhibit_time": "[f]",
                "acq_status": "[f]",
                "acq_time_of_flight": "[f]",
                "acq_start_point": "[f]",
                "sample_id": "[t]",
            }

        self.static_additions = {
            }

        self.key_translations = {
                "lat": ("[f]", "object_lat"),
                "latitude": ("[f]", "object_lat"),
                "lon": ("[f]", "object_lon"),
                "long": ("[f]", "object_lon"),
                "longitude": ("[f]", "object_lon"),
                "feature_extractor": ("[t]", "process_feature_extractor"),
                "volume_imaged": ("[f]", "acq_volimage"),
                "max_esd_save_threshold": ("[f]", "acq_min_esd"),
                "min_esd_save_threshold": ("[f]", "acq_max_esd"),
                "area": ("[f]", "object_area"),
                "convexarea": ("[f]", "object_convexarea"),
                "convexperimeter": ("[f]", "object_convexperimeter"),
                "extent": ("[f]", "object_extent"),
                "equivdiameter": ("[f]", "object_equivdiameter"),
                "eccentricity": ("[f]", "object_eccentricity"),
                "convexperimeter": ("[f]", "object_convexperimeter"),
                "convexarea": ("[f]", "object_convexarea"),
                "boundingbox_ywidth": ("[f]", "object_boundingbox_ywidth"),
                "boundingbox_xwidth": ("[f]", "object_boundingbox_xwidth"),
                "biovolume": ("[f]", "object_biovolume"),
                "bflip": ("[f]", "object_bflip"),
                "b90": ("[f]", "object_b90"),
                "b180": ("[f]", "object_b180"),
                "area_over_perimetersquared": ("[f]", "object_area_over_perimetersquared"),
                "area_over_perimeter": ("[f]", "object_area_over_perimeter"),
                "area": ("[f]", "object_area"),
                "date": ("[f]", "object_date"),
                "time": ("[f]", "object_time"),
                "texture_uniformity": ("[f]", "object_texture_uniformity"),
                "texture_third_moment": ("[f]", "object_texture_third_moment"),
                "texture_smoothness": ("[f]", "object_texture_smoothness"),
                "texture_entropy": ("[f]", "object_texture_entropy"),
                "texture_average_gray_level": ("[f]", "object_texture_average_gray_level"),
                "texture_average_contrast": ("[f]", "object_texture_average_contrast"),
                "surfacearea": ("[f]", "object_surfacearea"),
                "summedsurfacearea": ("[f]", "object_summedsurfacearea"),
                "summedperimeter": ("[f]", "object_summedperimeter"),
                "summedminoraxislength": ("[f]", "object_summedminoraxislength"),
                "summedmajoraxislength": ("[f]", "object_summedmajoraxislength"),
                "summedconvexperimeter_over_perimeter": ("[f]", "object_summedconvexperimeter_over_perimeter"),
                "summedconvexperimeter": ("[f]", "object_summedconvexperimeter"),
                "summedconvexarea": ("[f]", "object_summedconvexarea"),
                "summedbiovolume": ("[f]", "object_summedbiovolume"),
                "summedarea": ("[f]", "object_summedarea"),
                "solidity": ("[f]", "object_solidity"),
                "shapehist_skewness_normeqd": ("[f]", "object_shapehist_skewness_normeqd"),
                "shapehist_median_normeqd": ("[f]", "object_shapehist_median_normeqd"),
                "shapehist_mean_normeqd": ("[f]", "object_shapehist_mean_normeqd"),
                "shapehist_kurtosis_normeqd": ("[f]", "object_shapehist_kurtosis_normeqd"),
                "rwhalfpowerintegral": ("[f]", "object_rwhalfpowerintegral"),
                "rwcenter2total_powerratio": ("[f]", "object_rwcenter2total_powerratio"),
                "rotatedboundingbox_ywidth": ("[f]", "object_rotatedboundingbox_ywidth"),
                "rotatedboundingbox_xwidth": ("[f]", "object_rotatedboundingbox_xwidth"),
                "rotated_boundingbox_solidity": ("[f]", "object_rotated_boundingbox_solidity"),
                "roi_number": ("[f]", "object_roi_number"),
                "representativewidth": ("[f]", "object_representativewidth"),
                "perimeter": ("[f]", "object_perimeter"),
                "orientation": ("[f]", "object_orientation"),
                "numblobs": ("[f]", "object_numblobs"),
                "minoraxislength": ("[f]", "object_minoraxislength"),
                "minferetdiameter": ("[f]", "object_minferetdiameter"),
                "maxferetdiameter": ("[f]", "object_maxferetdiameter"),
                "majoraxislength": ("[f]", "object_majoraxislength"),
                "hflip_over_h180": ("[f]", "object_hflip_over_h180"),
                "hflip": ("[f]", "object_hflip"),
                "h90_over_hflip": ("[f]", "object_h90_over_hflip"),
                "h90_over_h180": ("[f]", "object_h90_over_h180"),
                "h90": ("[f]", "object_h90"),
                "h180": ("[f]", "object_h180"),
            }

        #,
        #        "": ("[]", "")

        if "ship_name" in self.options.keys():
            self.static_additions["sample_ship"] = self.options["ship_name"]
            self.ecotaxa_table_header["sample_ship"] = "[t]"
        if "cruise_name" in self.options.keys():
            self.static_additions["sample_cruise"] = self.options["cruise_name"]
            self.ecotaxa_table_header["sample_cruise"] = "[t]"
        if "project_name" in self.options.keys():
            self.static_additions["sample_project"] = self.options["project_name"]
            self.ecotaxa_table_header["sample_project"] = "[t]"
        if "station_id" in self.options.keys():
            self.static_additions["sample_stationid"] = self.options["station_id"]
            self.ecotaxa_table_header["sample_stationid"] = "[t]"
        if "ctd_cast" in self.options.keys():
            self.static_additions["sample_ctdcast"] = self.options["ctd_cast"]
            self.ecotaxa_table_header["sample_ctdcast"] = "[t]"
        if "sample_barcode" in self.options.keys():
            self.static_additions["sample_barcode"] = self.options["sample_barcode"]
            self.ecotaxa_table_header["sample_barcode"] = "[t]"
        if "sample_comment" in self.options.keys():
            self.static_additions["sample_comment"] = self.options["sample_comment"]
            self.ecotaxa_table_header["sample_comment"] = "[t]"
        if "sampling_gear" in self.options.keys():
            self.static_additions["sample_samplinggear"] = self.options["sampling_gear"]
            self.ecotaxa_table_header["sample_samplinggear"] = "[t]"
        if "initial_collected_volume_m3" in self.options.keys():
            self.static_additions["sample_initial_col_vol_m3"] = self.options["initial_collected_volume_m3"]
            self.ecotaxa_table_header["sample_initial_col_vol_m3"] = "[f]"
        if "concentrated_sample_volume_m3" in self.options.keys():
            self.static_additions["sample_concentrated_sample_volume"] = self.options["concentrated_sample_volume_m3"]
            self.ecotaxa_table_header["sample_concentrated_sample_volume"] = "[f]"
        if "dilution_factor" in self.options.keys():
            self.static_additions["sample_dilution_factor"] = self.options["dilution_factor"]
            self.ecotaxa_table_header["sample_dilution_factor"] = "[f]"
        if "operator_name" in self.options.keys():
            if not self.allow_short_names:
                if len(self.options["operator_name"]) < 5:
                    raise RuntimeError("Are you sure \"" + self.options["operator_name"] + "\" is a full name? To force allowing names shorter than 5 characters, use the option \"--reallyignoreshortnames\"")
            self.static_additions["sample_operator"] = self.options["operator_name"]
            self.ecotaxa_table_header["sample_operator"] = "[t]"
        if "dilution_method" in self.options.keys():
            self.static_additions["sample_dilution_method"] = self.options["dilution_method"]
            self.ecotaxa_table_header["sample_dilution_method"] = "[t]"
        if "fixative" in self.options.keys():
            self.static_additions["sample_fixative"] = self.options["fixative"]
            self.ecotaxa_table_header["sample_fixative"] = "[t]"
        if "sieve_min_um" in self.options.keys():
            self.static_additions["sample_sieve_min_um"] = self.options["sieve_min_um"]
            self.ecotaxa_table_header["sample_sieve_min_um"] = "[f]"
        if "sieve_max_um" in self.options.keys():
            self.static_additions["sample_sieve_max_um"] = self.options["sieve_max_um"]
            self.ecotaxa_table_header["sample_sieve_max_um"] = "[f]"
        if "feature_extractor_min_thresh" in self.options.keys():
            self.static_additions["process_min_thresh"] = self.options["feature_extractor_min_thresh"]
            self.ecotaxa_table_header["process_min_thresh"] = "[f]"
        if "feature_extractor_max_thresh" in self.options.keys():
            self.static_additions["process_max_thresh"] = self.options["feature_extractor_max_thresh"]
            self.ecotaxa_table_header["process_max_thresh"] = "[f]"
        if "um_per_pixel" in self.options.keys():
            self.static_additions["process_pixel_um"] = self.options["um_per_pixel"]
            self.ecotaxa_table_header["process_pixel_um"] = "[f]"
        if "sample_depth" in self.options.keys():
            self.static_additions["sample_depth_min"] = self.options["sample_depth"]
            self.ecotaxa_table_header["sample_depth_min"] = "[f]"
            self.static_additions["sample_depth_max"] = self.options["sample_depth"]
            self.ecotaxa_table_header["sample_depth_max"] = "[f]"
        if "meta_qc_by" in self.options.keys():
            if not self.allow_short_names:
                if len(self.options["meta_qc_by"]) < 5:
                    raise RuntimeError("Are you sure \"" + self.options["meta_qc_by"] + "\" is a full name? To force allowing names shorter than 5 characters, use the option \"--reallyignoreshortnames\"")
            self.static_additions["sample_meta_qc_by"] = self.options["meta_qc_by"]
            self.ecotaxa_table_header["sample_meta_qc_by"] = "[t]"



        if self.with_images:
            self.job_object.log_function("Creating EcoTaxa table for use alongside an image upload")
            self.ecotaxa_table_header["img_file_name"] = "[t]"
            self.ecotaxa_table_header["img_rank"] = "[f]"
        else:
            self.job_object.log_function("Creating EcoTaxa table only, without images")

        self.total_rois = 0
        for roi_reader in roi_readers:
            self.total_rois += len(roi_reader.rows)


    def add_csv_file(self, csv_fd, csv_filename):
        metadata_dict = list(csv.DictReader(csv_fd))
        if len(metadata_dict) > 0:
            dict_keys = metadata_dict[0].keys()
            can_join_dict = False
            metadata_dict_info = {
                    "data_keys": [],
                    "filename": csv_filename
                }
            bin_id_regex = re.compile(r'D[0-9]{8}T[0-9]{6}_IFCB[0-9]+')
            if ("bin" in dict_keys) or ("filename" in dict_keys):
                can_join_dict = True
                metadata_dict_info["bin_key_column"] = "filename"
                if "bin" in dict_keys:
                    metadata_dict_info["bin_key_column"] = "bin"
            if ("roi_number" in dict_keys):
                can_join_dict = True
                metadata_dict_info["roi_key_column"] = "roi_number"
                result = bin_id_regex.search(csv_filename)
                metadata_dict_info["bin_match"] = result.group()
                metadata_dict_info["roi_key_index"] = {}
                for row_index in range(len(metadata_dict)):
                    metadata_dict_info["roi_key_index"][int(metadata_dict[row_index][metadata_dict_info["roi_key_column"]])] = row_index
            if can_join_dict:
                metadata_dict_info["dict"] = metadata_dict
                matched_ecotaxa_columns = []
                matched_ecotaxa_types = []
                match_data_keys = []

                for candidate_key in self.key_translations.keys():
                    if candidate_key in dict_keys:
                        match_data_keys.append(candidate_key)
                        matched_ecotaxa_types.append(self.key_translations[candidate_key][0])
                        matched_ecotaxa_columns.append(self.key_translations[candidate_key][1])

                for column_index in range(len(matched_ecotaxa_columns)):
                    if matched_ecotaxa_columns[column_index] not in self.column_sources.keys():
                        self.column_sources[matched_ecotaxa_columns[column_index]] = []
                    self.ecotaxa_table_header[matched_ecotaxa_columns[column_index]] = matched_ecotaxa_types[column_index]
                    self.match_data_keys[matched_ecotaxa_columns[column_index]] = match_data_keys[column_index]
                    self.column_sources[matched_ecotaxa_columns[column_index]].append(metadata_dict_info)

                self.job_object.log_function("Loading metadata file \"" + csv_filename + "\" with " + str(len(metadata_dict_info["dict"])) + " rows")
        else:
            self.job_object.log_function("Skipping metadata file \"" + csv_filename + "\" with zero rows")

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.roi_readers[self.reader_index].rois):
            roi = self.roi_readers[self.reader_index].rois[self.index]
            dt = datetime.strptime(self.ifcb_ids[self.reader_index].split("_")[0], "D%Y%m%dT%H%M%S").replace(tzinfo=timezone.utc)
            observation_id = self.ifcb_ids[self.reader_index] + "_" + str(roi.index).zfill(5)
            ifcb_bin = self.ifcb_ids[self.reader_index]
            trigger_values = {}
            for key in roi.trigger.raw.keys():
                trigger_values[key] = float(roi.trigger.raw[key])
            trigger_id = self.ifcb_ids[self.reader_index] + "_TN" + str(int(trigger_values["trigger_number"]))
            ifcb_sn = self.ifcb_ids[self.reader_index].split("_")[1][4:]
            extents = [(0, roi.array.shape[0]), (0, roi.array.shape[1])]
            origin_extents = roi.array.shape

            record = {
                "object_id": observation_id,
                "object_date": dt.strftime("%Y-%m-%d"),
                "object_time":  dt.strftime("%H:%M:%S"),
                "object_roi_width": int(trigger_values["roi_width"]),
                "object_roi_height": int(trigger_values["roi_height"]),
                "process_id": ifcb_bin + "_SeaSTAR",
                "process_date": self.process_time.strftime("%Y-%m-%d"),
                "process_time": self.process_time.strftime("%H:%M:%S"),
                "acq_id": trigger_id,
                "acq_instrument": "IFCB",
                "acq_sn": ifcb_sn,
                "acq_operator": self.options["operator_name"],
                "acq_run_time": trigger_values["run_time"],
                "acq_trigger_number": trigger_values["trigger_number"],
                "acq_peak_b": trigger_values["peak_b"],
                "acq_pmt_b": trigger_values["pmt_b"],
                "acq_peak_a": trigger_values["peak_a"],
                "acq_pmt_a": trigger_values["pmt_a"],
                "acq_grab_time_end": trigger_values["grab_time_end"],
                "acq_grab_time_start": trigger_values["grab_time_start"],
                "acq_adc_time": trigger_values["adc_time"],
                "acq_signal_length": trigger_values["signal_length"],
                "acq_inhibit_time": trigger_values["inhibit_time"],
                "acq_status": trigger_values["status"],
                "acq_time_of_flight": trigger_values["time_of_flight"],
                "acq_start_point": trigger_values["start_point"],
                "sample_id": ifcb_bin,
            }

            # Add in denormalised data for the whole sample
            for key in self.static_additions.keys():
                record[key] = self.static_additions[key]

            # Add in data from CSV tables
            for ecotaxa_column in self.column_sources.keys():
                value = None
                for column_source in self.column_sources[ecotaxa_column]:
                    candidate = True
                    if "bin_match" in column_source.keys():
                        if not ifcb_bin == column_source["bin_match"]:
                            candidate = False
                    if candidate:
                        if "roi_key_column" in column_source.keys():
                            #for row in column_source["dict"]:
                            #    if int(row[column_source["roi_key_column"]]) == roi.index:
                            #        value = row[self.match_data_keys[ecotaxa_column]]
                            try:
                                row_index = column_source["roi_key_index"][roi.index]
                                value = column_source["dict"][row_index][self.match_data_keys[ecotaxa_column]]
                            except KeyError:
                                self.job_object.error_function("KeyError on matching ROI " + str(roi.index) + " from bin \"" + self.ifcb_ids[self.reader_index] + "\" to row in CSV \"" + column_source["filename"] + "\" for column \"" + ecotaxa_column + "\"")
                        if "bin_key_column" in column_source.keys():
                            for row in column_source["dict"]:
                                if row[column_source["bin_key_column"]] == ifcb_bin:
                                    value = row[self.match_data_keys[ecotaxa_column]]
                if value is "":
                    value = None # Treat empty data cells as null
                if value is None:
                    self.job_object.error_function("Missing data for ROI " + observation_id + " for column \"" + ecotaxa_column + "\"")
                    if self.mangle_nulls:
                        if self.ecotaxa_table_header[ecotaxa_column] == "[t]":
                            record[ecotaxa_column] = "None"
                        else:
                            record[ecotaxa_column] = "-99999999"

                else:
                    record[ecotaxa_column] = value

            # Only advance indexes at the end!
            self.index += 1
            if self.index >= len(self.roi_readers[self.reader_index].rois):
                if self.reader_index < (len(self.roi_readers) - 1):
                    self.reader_index += 1
                    self.index = 0

            if self.with_images:
                record["img_file_name"] = observation_id + ".png"
                record["img_rank"] = 0
                return (record, roi.image)
            else:
                return (record, )
        raise StopIteration

class MainJob:
    def calc_progress_report(self):
        self.last_time = time.time()
        proportion = self.currently_processing/self.total_rois
        elapsed_time = self.last_time - self.first_time
        time_per_roi = elapsed_time / self.currently_processing
        remaining_rois = self.total_rois - self.currently_processing
        remaining_time = time_per_roi * remaining_rois
        self.report_progress(proportion, remaining_time)

    def __init__(self, options, progress_reporting_function = lambda prop, etr : print(str(int(prop*10000)/100) + "% done - ETR " + str(int(etr)) + "s"), log_function = lambda txt : print("[LOG] " + txt), error_function = lambda txt, err_level : print("[ERR] " + txt)):

        self.options = options
        self.report_progress = progress_reporting_function
        self.log_function = log_function
        self.error_function = error_function

        #print(options)

        if "table_only" in options.keys():
            self.with_images = not options["table_only"]
        else:
            self.with_images = True

        input_files_list = []
        for input_file_path in options["input_files"]:
            input_files_list.append(os.path.abspath(input_file_path))

        #print(input_files_list)

        ifcb_bins = []
        ifcb_files = []
        roi_readers = []

        intermediate_files_list = set()
        csv_files_list = set()
        for file_name in input_files_list:
            splitext = os.path.splitext(file_name)
            if (splitext[1] == ".hdr") or (splitext[1] == ".adc") or (splitext[1] == ".roi"):
                intermediate_files_list.add(splitext[0])
            elif (splitext[1] == ".csv"):
                csv_files_list.add(file_name)

        for file_name in intermediate_files_list:
            ifcb_bins.append(os.path.basename(file_name))
            ifcb_files.append(file_name)

        if len(ifcb_bins) == 0:
            raise RuntimeException("No IFCB bins supplied!")

        self.log_function("Creating ROIReaders")

        for i in range(len(ifcb_files)):
            self.log_function("Loading \"" + ifcb_files[i] + ".hdr\"")
            roi_readers.append(libifcb.ROIReader(ifcb_files[i] + ".hdr", ifcb_files[i] + ".adc", ifcb_files[i] + ".roi"))

        self.log_function("Initialising entry provider")
        self.entry_provider = IFCBEntryProvider(roi_readers, ifcb_bins, self, self.with_images, options)

        for csv_file in csv_files_list:
            self.entry_provider.add_csv_file(open(csv_file, "r"), csv_file)

        self.total_rois = self.entry_provider.total_rois

    def execute(self):
        self.first_time = time.time()
        self.currently_processing = 0
        self.transferred_rois = 0
        self.last_time = time.time()

        os.makedirs(os.path.dirname(self.options["output_file"]), exist_ok=True)

        if self.with_images:
            zip_files = 1
            out_zip = None
            max_size = None
            out_file = self.options["output_file"]
            out_file_se = os.path.splitext(out_file)
            if max_size is not None:
                out_zip = zipfile.ZipFile(out_file_se[0] + "_part" + str(zip_files) + out_file_se[1], 'w')
            else:
                out_zip = zipfile.ZipFile(out_file, 'w')

            container = os.path.basename(out_file_se[0]) + "_part" + str(zip_files)

            ecotaxa_md = io.StringIO()
            ecotaxa_md_writer = csv.DictWriter(ecotaxa_md, fieldnames=self.entry_provider.ecotaxa_table_header.keys(), quoting=csv.QUOTE_NONNUMERIC, delimiter='\t', lineterminator='\n')
            ecotaxa_md_writer.writeheader()
            ecotaxa_md_writer.writerow(self.entry_provider.ecotaxa_table_header)
            running_compressed_size = 0
            tsv_name_suffix = ""

            try:
                while True:
                    entry = next(self.entry_provider)
                    self.currently_processing += 1
                    tsv_name_suffix = entry[0]["sample_id"] # So every TSV is likely to have a different, but predictable name
                    ecotaxa_md_writer.writerow(entry[0])

                    imbuffer = io.BytesIO()
                    entry[1].save(imbuffer, "png")
                    imbytes = imbuffer.getvalue()
                    running_compressed_size += len(imbytes)
                    out_zip.writestr(container + "/" + entry[0]["img_file_name"], imbytes)
                    self.transferred_rois += 1

                    if (self.currently_processing % 512) == 0:
                        self.calc_progress_report()

                    if max_size is not None:
                        if running_compressed_size > max_size:
                            zip_files += 1
                            out_zip.writestr(container + "/ecotaxa_" + tsv_name_suffix + ".tsv", ecotaxa_md.getvalue())
                            out_zip.close()
                            out_zip = zipfile.ZipFile(out_file_se[0] + "_part" + str(zip_files) + out_file_se[1], 'w')
                            container = os.path.basename(out_file_se[0]) + "_part" + str(zip_files)
                            ecotaxa_md = io.StringIO()
                            ecotaxa_md_writer = csv.DictWriter(ecotaxa_md, fieldnames=self.entry_provider.ecotaxa_table_header.keys(), quoting=csv.QUOTE_NONNUMERIC, delimiter='\t', lineterminator='\n')
                            ecotaxa_md_writer.writeheader()
                            ecotaxa_md_writer.writerow(self.entry_provider.ecotaxa_table_header)
                            running_compressed_size = 0
            except StopIteration:
                pass

            out_zip.writestr(container + "/ecotaxa_" + tsv_name_suffix + ".tsv", ecotaxa_md.getvalue())
            out_zip.close()

            self.log_function("Saved " + str(self.transferred_rois) + " ROIs to EcoTaxa table")
        else:
            with open(self.options["output_file"], "w") as ecotaxa_md:
                ecotaxa_md_writer = csv.DictWriter(ecotaxa_md, fieldnames=self.entry_provider.ecotaxa_table_header.keys(), quoting=csv.QUOTE_NONNUMERIC, delimiter='\t', lineterminator='\n')
                ecotaxa_md_writer.writeheader()
                ecotaxa_md_writer.writerow(self.entry_provider.ecotaxa_table_header)

                try:
                    while True:
                        entry = next(self.entry_provider)
                        self.currently_processing += 1
                        ecotaxa_md_writer.writerow(entry[0])
                        self.transferred_rois += 1
                        if (self.currently_processing % 512) == 0:
                            self.calc_progress_report()
                except StopIteration:
                    pass

            self.log_function("Saved " + str(self.transferred_rois) + " ROIs to EcoTaxa table")





import sys
import importlib
import os
import json
import time
import math
import textwrap
import datetime
import glob

long_help_text = """
seastar --gui                           Launch the SeaSTAR GUI
seastar <job_name> [flags]              General structure

With any SeaSTAR job you can use --logfile <file_path> to send all log output
to a file in addition to the console. This log output will also include the date
and time, along with all arguments used to invoke SeaSTAR.
"""

def base_cli():
    global last_line_terminated
    last_line_terminated = True

    python_file_loc = os.path.dirname(os.path.realpath(__file__))

    job_dirs = os.scandir(os.path.join(python_file_loc, "jobs"))
    #found_job_modules = {}
    module_io_defs = {}
    for job_dir in job_dirs:
        if job_dir.is_dir():
            if not job_dir.name.startswith("__"):
                full_job_dir = os.path.join(python_file_loc, "jobs", job_dir.name)
                with open(os.path.join(full_job_dir, "io.json"), "r") as io_json_fp:
                    module_io_defs[job_dir.name] = json.loads(io_json_fp.read())
                    #print("loading jobs." + job_dir.name)
                    module_io_defs[job_dir.name]["job_module"] = "seastartool.jobs." + job_dir.name
                    #options = {}
                    #found_job_modules[job_dir.name].MainJob(options)



    #for
    #found_job_modules[module_name] = importlib.import_module(module_path)

    eargs = sys.argv[1:]
    help_flag = False
    explicit_help_flag = False
    ehelp_msg = "No command specified"


    mode = "command"
    mode_stack = []
    capture_heap = []
    command = None
    capture_option = None
    multiple_capture_switch = False
    options = {}
    io_def = None
    gui_flag = False
    logfile_path = None

    for arg in eargs:
        if arg.startswith("--"):
            if multiple_capture_switch:
                multiple_capture_switch = False
                mode = mode_stack.pop() # Break out of the current multiple capture
            if arg == "--help":
                ehelp_msg = None
                help_flag = True
                explicit_help_flag = True
                break
            if arg == "--gui":
                gui_flag = True
            elif arg == "--logfile":
                mode_stack.append(mode)
                mode = "logfile_capture"
            else:
                option_recognised = False
                if io_def is not None:
                    for io_def_key in io_def["inputs"].keys():
                        if "cli_arg" in io_def["inputs"][io_def_key].keys():
                            if io_def["inputs"][io_def_key]["cli_arg"] == arg[2:]:
                                capture_option = io_def_key
                                if "multiple" in io_def["inputs"][io_def_key].keys():
                                    multiple_capture_switch = io_def["inputs"][io_def_key]["multiple"]
                                if io_def["inputs"][io_def_key]["type"] == "BOOLEAN":
                                    options[io_def_key] = True
                                else:
                                    mode_stack.append(mode)
                                    if multiple_capture_switch:
                                        mode = "multi_capture"
                                    else:
                                        mode = "single_capture"
                                option_recognised = True
                                break

                if not option_recognised:
                    ehelp_msg = "Unrecognised option \"" + arg + "\""
                    help_flag = True
                    break
        elif arg.startswith("-"):
            if multiple_capture_switch:
                multiple_capture_switch = False
                mode = mode_stack.pop() # Break out of the current multiple capture
            if arg == "-h":
                ehelp_msg = None
                help_flag = True
                explicit_help_flag = True
                break
            else:
                option_recognised = False
                if io_def is not None:
                    for io_def_key in io_def["inputs"].keys():
                        if "cli_short" in io_def["inputs"][io_def_key].keys():
                            if io_def["inputs"][io_def_key]["cli_short"] == arg[1:]:
                                capture_option = io_def_key
                                if "multiple" in io_def["inputs"][io_def_key].keys():
                                    multiple_capture_switch = io_def["inputs"][io_def_key]["multiple"]
                                if io_def["inputs"][io_def_key]["type"] == "BOOLEAN":
                                    options[io_def_key] = True
                                else:
                                    mode_stack.append(mode)
                                    if multiple_capture_switch:
                                        mode = "multi_capture"
                                    else:
                                        mode = "single_capture"
                                option_recognised = True
                                break

                if not option_recognised:
                    ehelp_msg = "Unrecognised option \"" + arg + "\""
                    help_flag = True
                    break
        else:
            if mode == "command":
                if arg == "help":
                    ehelp_msg = None
                    help_flag = True
                    explicit_help_flag = True
                    break
                elif arg == "gui":
                    gui_flag = True
                    break
                else:
                    if arg in module_io_defs.keys():
                        io_def = module_io_defs[arg]
                        command = arg
                        ehelp_msg = None
                    else:
                        ehelp_msg = "Unrecognised command \"" + arg + "\""
                        help_flag = True
                        break
            elif mode == "multi_capture":
                capture_heap.append(arg)
                options[capture_option] = capture_heap
            elif mode == "single_capture":
                options[capture_option] = arg
                mode = mode_stack.pop()
            elif mode == "logfile_capture":
                logfile_path = arg
                mode = mode_stack.pop()

    if command is None:
        if not gui_flag:
            help_flag = True

    if not help_flag:
        if io_def is not None:
            for option_key in io_def["inputs"].keys():
                if option_key not in options.keys():
                    if "default" in io_def["inputs"][option_key].keys():
                        options[option_key] = io_def["inputs"][option_key]["default"]
                    elif "required" in io_def["inputs"][option_key].keys():
                        if io_def["inputs"][option_key]["required"]:
                            help_flag = True
                            ehelp_msg = "Missing required option \"" + option_key +  "\", specified by"
                            if "cli_arg" in io_def["inputs"][option_key]:
                                ehelp_msg += " \"--" + io_def["inputs"][option_key]["cli_arg"] + " <value>\""
                            if "cli_short" in io_def["inputs"][option_key]:
                                ehelp_msg += " \"-" + io_def["inputs"][option_key]["cli_short"] + " <value>\""
                            if "hint" in io_def["inputs"][option_key].keys():
                                ehelp_msg += "\nHint: " + io_def["inputs"][option_key]["hint"]

    if help_flag:
        print("")
        print("SeaSTAR")
        print("Sea-faring System for Tagging, Attribution and Redistribution")
        print("")
        print("Copyright 2025, A Baldwin <alewin@noc.ac.uk>, National Oceanography Centre")
        print("This program comes with ABSOLUTELY NO WARRANTY. This is free software,")
        print("and you are welcome to redistribute it under the conditions of the")
        print("GPL version 3 license.")
        print("")
        if ehelp_msg is not None:
            if not explicit_help_flag:
                print("ERROR")
                print(ehelp_msg)
                print("")
        if (command == "help") or (command is None):
            print("Common usage:")
            print(long_help_text)
            print("Use \"seastar <job_name> --help\" to get specific help for a given job.")
            print("The following jobs are avaliable:")
            print("")
            for command_name in module_io_defs.keys():
                title = module_io_defs[command_name]["name"]
                description = module_io_defs[command_name]["description"]
                print(command_name.ljust(40, " ") + title)
                print("")
                text_wrapper = textwrap.TextWrapper(width=79-4)
                print("    " + text_wrapper.fill(text=description).replace("\n", "\n    "))
                print("")
        else:

            io_def = module_io_defs[command]
            title = io_def["name"]
            description = io_def["description"]
            print(command.ljust(40, " ") + title)
            print("")
            text_wrapper = textwrap.TextWrapper(width=79)
            print(text_wrapper.fill(text=description))
            print("")

            text_wrapper = textwrap.TextWrapper(width=79-40-1)

            for input_def_key in io_def["inputs"]:

                input_def = io_def["inputs"][input_def_key]
                input_sig = "      "
                input_tail_lines = ""
                semantic_type = "TEXT"
                if "type" in input_def.keys():
                    if input_def["type"] == "BOOLEAN":
                        semantic_type = "SWITCH"
                if "semantic_type" in input_def.keys():
                    semantic_type = input_def["semantic_type"]
                input_sig_ed = "<text>"
                if semantic_type == "MULTIPLE_FILES":
                    input_sig_ed = "<file> [file [...]]"
                if semantic_type == "SINGLE_FILE":
                    input_sig_ed = "<file>"
                elif semantic_type == "SINGLE_FOLDER":
                    input_sig_ed = "<folder>"
                elif semantic_type == "SWITCH":
                    input_sig_ed = ""

                hint_text = input_def["hint"]

                if "cli_short" in input_def.keys():
                    input_sig = "  -" + input_def["cli_short"] + ", "
                if "cli_arg" in input_def.keys():
                    input_sig += "--" + input_def["cli_arg"]
                if "required" in input_def.keys():
                    if input_def["required"]:
                        hint_text = "[REQUIRED] " + hint_text

                input_sig = input_sig + " " + input_sig_ed

                if "hint" in input_def.keys():
                    input_tail_lines = text_wrapper.fill(hint_text).replace("\n", "\n" + (" " * 40))
                input_str_def = input_sig.ljust(40, " ") + input_tail_lines
                print(input_str_def)
    else:
        for option_key in options.keys():
            input_def = module_io_defs[command]["inputs"][option_key]
            if "semantic_type" in input_def.keys():
                #print(option_key + " is " + input_def["semantic_type"])
                if input_def["semantic_type"] == "MULTIPLE_FILES":
                    if "multiple" in input_def.keys():
                        if input_def["multiple"]:
                            new_option_arr = []
                            for input_str in options[option_key]:
                                new_option_arr = new_option_arr + glob.glob(input_str)

                            #print("GLOB RECONSTRUCTION TRIGGER")
                            #print(options[option_key])
                            #print(new_option_arr)

                            options[option_key] = new_option_arr




        if gui_flag:
            from .gui import SeaSTARGUI # Avoid loading the GUI if the user doesn't want it!
            gui = SeaSTARGUI(python_file_loc=python_file_loc, module_io_defs=module_io_defs)

            gui.render_start_page()
            gui.enter_mainloop()
        else:

            job_start_time = time.time()

            logfile_fh = None

            if logfile_path is not None:
                logfile_fh = open(logfile_path, "a")
                curr_datetime = datetime.datetime.now(datetime.timezone.utc) # Do not use datetime.UTC - only avail on Python 3.11 or newer
                logfile_fh.write("[LOG - " + curr_datetime.strftime("%Y-%m-%d %H:%M:%S") + "] SeaSTAR invoked with following command\n")
                logfile_fh.write("[LOG - " + curr_datetime.strftime("%Y-%m-%d %H:%M:%S") + "] " + " ".join(sys.argv[1:]) + "\n")

            def prf(prop, etr):
                bar_w = 16
                bar_x = round(bar_w * prop)
                bar_l = "#"*bar_x
                bar_r = "_"*(bar_w - bar_x)
                percent = f"{prop:.2%}"
                secs = round(etr)
                timestr = f"about {secs}s remaining..."
                if secs < 3:
                    timestr = "only a few seconds remaining..."
                if secs > 60:
                    mins = math.floor(secs / 60)
                    secs = secs - (mins * 60)
                    timestr = f"about {mins}min {secs}s remaining..."
                if secs > 3600:
                    hrs = math.floor(mins / 60)
                    mins = mins - (hrs * 60)
                    timestr = f"about {hrs}hr {mins}min remaining..."

                print(f"\r[{bar_l}{bar_r}] {percent} done, {timestr}".ljust(79, " "), end="")

            def log_function(txt):
                if last_line_terminated:
                    print("[LOG] " + txt)
                else:
                    print("\n[LOG] " + txt, end="")
                if logfile_fh is not None:
                    curr_datetime = datetime.datetime.now(datetime.timezone.utc)
                    logfile_fh.write("[LOG - " + curr_datetime.strftime("%Y-%m-%d %H:%M:%S") + "] " + txt + "\n")

            def error_function(txt, err_level):
                if last_line_terminated:
                    print("[ERR] " + txt)
                else:
                    print("\n[ERR] " + txt, end="")
                if logfile_fh is not None:
                    curr_datetime = datetime.datetime.now(datetime.timezone.utc)
                    logfile_fh.write("[ERR - " + curr_datetime.strftime("%Y-%m-%d %H:%M:%S") + "] " + txt + "\n")

            log_function("Preparing job...")

            main_job_object = importlib.import_module(module_io_defs[command]["job_module"]).MainJob(options, prf, log_function, error_function)
            log_function("Starting main processing script...")
            last_line_terminated = False
            main_job_object.execute()
            job_end_time = time.time()

            secs = round(job_end_time - job_start_time)
            timestr = f"{secs}s"
            if secs > 60:
                mins = math.floor(secs / 60)
                secs = secs - (mins * 60)
                timestr = f"{mins}min {secs}s"
            if secs > 3600:
                hrs = math.floor(mins / 60)
                mins = mins - (hrs * 60)
                timestr = f"{hrs}hr {mins}min"

            log_function(f"Finished in  {timestr}".ljust(79, " "))
            last_line_terminated = True
            print("")
            log_function("Done!")

            if logfile_fh is not None:
                logfile_fh.close()

